/*
 *  asmconv.c - Z80 assembler output to 83p
 *  Copyright (C) 1998
 *
 *  Original Author: Steven Fuller <relnev@netscape.net>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <getopt.h>

#define NAME_LENGTH 8
#define COMMENT_LENGTH 0x2A

void usage(char *s)
{
	printf("\n");
	printf("%s - Z80 assembler output to 83p\n", s);
	printf("by Steven Fuller <relnev@netscape.net>\n");
	printf("\tusage: %s <filename>\n", s);
	printf("Optional parameters:\n");
	printf("\t-b: filename is in binary format\n");
	printf("\t-h: this text\n");
	printf("\t-n <name>: program name\n");
	printf("\t-o <file>: output filename\n");
	printf("\t-p: protect\n");
	printf("\t-s: <comment>: 83p file comment\n");
	printf("\t-z: output in 'squished' format\n");
	printf("\n");
}

int main(int argc, char *argv[])
{
	int opt, x, filelen, inlen, outlen;
	/* int calctype; */
	int protected = 0;
	int squished = 0;
	int binary = 0;
	unsigned char *buffer, *filebuf;
	unsigned char *name, *fileinfo;
	unsigned short chksum, totlen;
	char *infile, *outfile;
	FILE *infp, *outfp;	
	unsigned char sig[] = {'*', '*', 'T', 'I', '8', '3', '*', '*', 0x1A, 0x0A, 0x0};
	unsigned char endstr[] = {'?', 212, '?', '0', '0', '0', '0', '?', 212};
	char convstr[] = {0, 0, 0};
	
	infile = outfile = NULL;
	name = fileinfo = NULL;
	
/*	
	if (!isatty(fileno(stdout))) {
		fprintf(stderr, "You are not a terminal!\n");
		exit(EXIT_FAILURE);
	}
*/
	
	while ((opt = getopt(argc, argv, "bc:hn:o:ps:z")) != -1) {
		switch(opt) {
		case 'b':
			binary = 1;
			break;
		case 'c':
			printf("not done yet!\n");
			exit(EXIT_FAILURE);
		case 'h':
			usage(argv[0]);
			exit(EXIT_SUCCESS);
		case 'n':
			x = strlen(optarg);
			if (x > 8) {
				fprintf(stderr, "Name \"%s\" is too long!\n", optarg);
				exit(EXIT_FAILURE);
			}
			name = (unsigned char *)malloc(NAME_LENGTH);
			memset(name, '\0', NAME_LENGTH);
			strncpy((char *)name, optarg, NAME_LENGTH);
			if (!isprint(name[0])) {
				fprintf(stderr, "Name \"%s\" is invalid!\n", optarg);
				exit(EXIT_FAILURE);
			}
			name[0] = toupper(name[0]);
			for (x = 1; x < strlen((char *)name); x++) {
				if (!isalnum(name[x])) {
					fprintf(stderr, "Name \"%s\" is invalid!\n", optarg);
					exit(EXIT_FAILURE);
				}
				name[x] = toupper(name[x]);
			}
			break;
		case 'o':
			outfile = (char *)strdup(optarg);
			break;
		case 'p':
			protected = 1;
			break;
		case 's':
			x = strlen(optarg);
			if (x > COMMENT_LENGTH) {
				fprintf(stderr, "File info \"%s\" is too long!\n", optarg);
				exit(EXIT_FAILURE);
			}
			fileinfo = (unsigned char *)malloc(COMMENT_LENGTH + 1);
			memset(fileinfo, '\0', COMMENT_LENGTH + 1);
			strcpy((char *)fileinfo, optarg);
			for (x = 0; x < strlen((char *)fileinfo); x++) {
				if (!(isalnum(fileinfo[x]) || (fileinfo[x] == ' '))) {
					fprintf(stderr, "File info \"%s\" is invalid!\n", optarg);
					exit(EXIT_FAILURE);
				}
			}
			break;
		case 'z':
			squished = 1;
			break;
		case ':':
			fprintf(stderr, "option needs a value\n");
			exit(EXIT_FAILURE);
		case '?':
			fprintf(stderr, "unknown option: %c\n", optopt);
			exit(EXIT_FAILURE);
		}
	}
	
	if ((argc - optind) != 1) {
		fprintf(stderr, "Incorrect amount of parameters, Try %s -h\n", argv[0]);
		exit(EXIT_FAILURE);
	}
	
	infile = (char *)strdup(argv[optind]);
	
	if (!(infp = fopen(infile, "rb"))) {
		fprintf(stderr, "Unable to open inputfile \"%s\".\n", infile);
		exit(EXIT_FAILURE);
	}
	
	if (outfile == NULL) {
		for (x = strlen(infile) - 1; x >= 0; x--)
			if (*(infile + x) == '.')
				break;
		if (x < 0) x = strlen(infile);
		outfile = (char *)malloc(x + 5);
		memset(outfile, '\0', x + 5);
		strncpy(outfile, infile, x);
		strcat(outfile, ".83p");
		/* printf("Outfile -> %s\n", outfile); */
	}
	
	if (!(outfp = fopen(outfile, "wb"))) {
		fprintf(stderr, "Unable to open outputfile \"%s\".\n", outfile);
		exit(EXIT_FAILURE);
	}
	
	printf("Outfile -> %s\n", outfile);
	
	fseek(infp, 0, SEEK_END);
	filelen = ftell(infp);
	rewind(infp);
	
	if (binary) {
		buffer = (unsigned char *)malloc(filelen);
		memset(buffer, '\0', filelen);
		fread(buffer, filelen, 1, infp);
		
		fclose(infp);
		if (squished) {
			filebuf = (unsigned char *)malloc(filelen);
			memcpy(filebuf, buffer, filelen);
			outlen = filelen;
		} else {
			filebuf = (unsigned char *)malloc(filelen * 2);
			for (x = 0; x < filelen; x++) 
				sprintf((char *)(filebuf + (x * 2)), "%02X", *(buffer + x));     
			outlen = filelen * 2;
		}
	} else {
		buffer = (unsigned char *)malloc(filelen + 1);
		memset(buffer, '\0', filelen + 1);
		fread(buffer, filelen, 1, infp);
		fclose(infp);
	
		filebuf = (unsigned char *)malloc(filelen);
		memset(filebuf, '\0', filelen + 1);
		inlen = outlen = 0;
		for (;;) {
			if ((inlen + 12) == filelen) {
				if (strncmp((char *)(buffer + inlen), ":00000001FF\n", 12) == 0)
					break;	
			} else if ((inlen + 12) > filelen) { 
				fprintf(stderr, "bad input filelength!\n"); 
				exit(EXIT_FAILURE);
			}
			if ((inlen + 9) < filelen) {
				inlen += 9;
				while ((*(buffer + inlen) != '\n') && (inlen < filelen)) {
					if (squished) {
						convstr[0] = *(buffer + inlen);
						convstr[1] = *(buffer + inlen + 1);
						*(filebuf + outlen) = (unsigned char)strtol(convstr, NULL, 16);
						outlen++;
						inlen += 2;
					} else { 
						*(filebuf + outlen) = *(buffer + inlen);
						outlen++;
						inlen++;
					}
				}
				if (squished)
					outlen--;
				else
					outlen -= 2;
				inlen++;
			} else {
				fprintf(stderr, "bad input filelength!\n");
				exit(EXIT_FAILURE);
			}
		}
		
		if (!(filebuf = realloc(filebuf, outlen + 1))) {
			fprintf(stderr, "realloc failed?? what!\n");
			exit(EXIT_FAILURE);
		}
		free(buffer);
	}
	
	totlen = (unsigned short)outlen + sizeof(endstr) + 2;

	fwrite(sig, sizeof(sig), 1, outfp);
	
	if (fileinfo) {
		fwrite(fileinfo, COMMENT_LENGTH, 1, outfp);
	} else {
		for (x = 0; x < COMMENT_LENGTH; x++)
			fputc(0x0, outfp);
	}
	
	if (!name) {
		name = (unsigned char *)malloc(NAME_LENGTH);
		memset(name, '\0', NAME_LENGTH);
		strcpy((char *)name, "TEST");
	}
	
	chksum = totlen + 7 + NAME_LENGTH;
	fputc((chksum & 0x00FF), outfp);
	fputc((chksum >> 8), outfp);
	
	chksum = 0;
	
	x = NAME_LENGTH + 3;
	fputc((x & 0x00FF), outfp); chksum += (x & 0x00FF);
	fputc((x >> 8), outfp); chksum += (x >> 8);
	
	fputc((totlen & 0x00FF), outfp); chksum += (totlen & 0x00FF);
	fputc((totlen >> 8), outfp); chksum += (totlen >> 8);
	
	if (protected) {
		fputc(0x6, outfp); chksum += 0x6;
	} else {
		fputc(0x5, outfp); chksum += 0x5;
	}
	
	for (x = 0; x < NAME_LENGTH; x++) {
		fputc(name[x], outfp);
		chksum += name[x];
	}	

	fputc((totlen & 0x00FF), outfp); chksum += (totlen & 0x00FF);
	fputc((totlen >> 8), outfp); chksum += (totlen >> 8);
	
	fputc(((totlen - 2) & 0x00FF), outfp); chksum += ((totlen - 2) & 0x00FF);
	fputc(((totlen - 2) >> 8), outfp); chksum += ((totlen - 2) >> 8);
	for (x = 0; x < outlen; x++) {
		fputc(*(filebuf + x), outfp);
		chksum += *(filebuf + x);
	}
	
	for (x = 0; x < sizeof(endstr); x++) {
		fputc(endstr[x], outfp);
		chksum += endstr[x];
	}
	
	fputc((chksum & 0x00FF), outfp);
	fputc((chksum >> 8), outfp);
	
	fclose(outfp);
	free(filebuf);
	free(outfile);
	free(infile);
	free(name);
	if (fileinfo) free(fileinfo);
	exit(EXIT_SUCCESS);
}
