/*
 *  group.c - grouping individual 83 files into .83g
 *  Copyright (C) 1998
 *
 *  Original Author: Steven Fuller <relnev@netscape.net>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>

#define MODE_GROUP	0
#define MODE_UNGROUP	1

void usage(char *s)
{
	printf("\n");
	printf("%s - group/ungroup manipulation of .83g files\n", s);
	printf("by Steven Fuller <relnev@netscape.net>\n");
	printf("\tusage: %s <groupfile to create> <files to add>\n", s);
	printf("Optional parameters:\n");
	printf("\t-h: this text\n");
	printf("\t-u: ungroup mode (not implemented yet)\n");
	printf("\t\tnew usage: %s -s <groupfile to ungroup>\n", s);
	printf("\n");
}

int main(int argc, char *argv[])
{
	FILE *infp, *outfp;
	unsigned char *buf;
	unsigned int length;
	unsigned int len;
	int mode = 0;
	
	unsigned short int chksum;
	
	signed char sig[] = {'*', '*', 'T', 'I', '8', '3', '*', '*', 0x1A, 0x0A, 0x0};
	
	int x, y;
	int opt;
	
	while ((opt = getopt(argc, argv, "hs:u")) != -1) {	
		switch(opt) {
			case 'h':
				usage(argv[0]);
				exit(EXIT_SUCCESS);
			case 's':
				break;
			case 'u':
				mode = MODE_UNGROUP;
				break;
			case ':':
				fprintf(stderr, "option needs a value\n");
				exit(EXIT_FAILURE);
			case '?':
				fprintf(stderr, "unknown option: %c\n", optopt);
				exit(EXIT_FAILURE);
		}
	}
	
	if (mode == MODE_UNGROUP) {
		if ((argc - optind) != 1) {
			fprintf(stderr, "Incorrect amount of parameters, Try %s -h\n", argv[0]);
			exit(EXIT_FAILURE);
		}
		printf("Not Done Yet\n");
		exit(EXIT_SUCCESS);
	}
	
	/* mode == MODE_GROUP */
	
	if ((argc - optind) < 2) {
		fprintf(stderr, "Incorrect amount of parameters, Try %s -h\n", argv[0]);
		exit(EXIT_FAILURE);
	}
	
	outfp = fopen(argv[optind], "w+");
	
	fwrite(sig, sizeof(sig), 1, outfp);
	
	for (x = 0; x < 0x2A; x++)
		fputc(0x0, outfp);
	
	fputc(0x0, outfp); /* placeholder */
	fputc(0x0, outfp); 
	
	chksum = 0;
	length = 0;
	
	for (x = optind + 1; x < argc; x++) {
		if ((infp = fopen(argv[x], "r")) == NULL) {
			printf("Unable to open %s...\n", argv[x]);
			continue;
		}
		
		fseek(infp, 0, SEEK_END);
		len = ftell(infp);
		rewind(infp);
		
		buf = (unsigned char *)malloc(len);
		fread(buf, len, 1, infp);

		fclose(infp);
		
		if (strncmp(sig, buf, sizeof(sig)) != 0) {
			printf("%s is not a TI-83 File!  Skipping...\n", argv[x]);
			free(buf);
			continue;
		}
				
		fwrite(buf + 55, len - 57, 1, outfp);
		length += len - 57;
		for (y = 55; y < len - 2; y++) 
			chksum += *(buf + y);
		
		free(buf);	
	}
	
	fputc((chksum & 0x00FF), outfp);
	fputc((chksum >> 8), outfp);
	
	fseek(outfp, 53, SEEK_SET);
	fputc((length & 0x00FF), outfp);
	fputc((length >> 8), outfp);
	
	fclose(outfp);
	return 0;
}
 