/*
 *  piconv.c - small picture converter
 *  Copyright (C) 1998
 *
 *  Original Author: Steven Fuller <relnev@netscape.net>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>
#include <pbm.h>

#define COMMENT_LENGTH 0x2A

void usage(char *s)
{
	printf("\n");
	printf("%s - small picture converter\n", s);
	printf("by Steven Fuller <relnev@netscape.net>\n");
	printf("\tusage: piconv <old.file> <new.file>\n");
	printf("Conversions supported:\n");
	printf("\t83i -> pbm\n");
	printf("\t83i -> inc\n");
	printf("\tpbm -> 83i\n");
	printf("\tpbm -> inc\n");
	printf("\n");
}

int pic_to_pbm(char *pic_file, char *pbm_file)
{
	FILE *fp;
	
	bit **pbm_data;

	int x, y, z;
	unsigned char *ptr;
	unsigned char *pic;
	int len, len2;
	
	fp = fopen(pic_file, "r");
	fseek(fp, 0, SEEK_END);
	len = ftell(fp);
	rewind(fp);
	
	pic = (unsigned char *)malloc(len);	
	fread(pic, len, 1, fp);
	fclose(fp);
		
	if (strncmp("**TI83**", (char *)pic, 8) != 0) {
		printf("Invalid TI83 File!\n");
		return 0;
	}
	
	ptr = pic;
	ptr += 11; /* file type ? */
	ptr += 0x2A; /* comment block */
	
	len2 = *ptr | (*(ptr + 1) << 8);
	if (len != (len2 + 57)) {
		printf("Multiple files detected! Only first one can be used...\n");
	}
		
	ptr += 2; /* file length */
	ptr += 2; /* 11 */
	ptr += 2; /* length of data */
	
	if (*ptr != 0x7) {
		printf("File is not a TI83 Pic!\n");
		return 0;
	}
	
	ptr += 1; /* type */ 
	ptr += 8; /* name */
	ptr += 2; /* length again ? */
	ptr += 2; /* length - 2 ? */
	
	pbm_data = (bit **)malloc(sizeof(bit *) * 63);
	for (x = 0; x < 63; x++) {
		pbm_data[x] = (bit *)calloc(sizeof(bit *) * 96, 1);
		for (y = 0; y < (96 / 8); y++) {
			for (z = 0; z < 8; z++)
				pbm_data[x][(y * 8) + z] = (*ptr & (1 << (7 - z))) >> (7 - z);
			ptr++;
		}
	}
	
	fp = fopen(pbm_file, "wb");
	pbm_writepbm(fp, (bit **)pbm_data, 96, 63, 0);
	fclose(fp);
	
	for (x = 0; x < 63; x++)
		free(pbm_data[x]);
	free(pbm_data);
	free(pic);
	
	return 1;		
}

int pic_to_inc(char *pic_file, char *inc_file)
{
	FILE *fp;

	int x, y;
	unsigned char *ptr;
	unsigned char *pic;
	int len, len2;
                        	
	fp = fopen(pic_file, "r");
	fseek(fp, 0, SEEK_END);
	len = ftell(fp);
	rewind(fp);
	
	pic = (unsigned char *)malloc(len);	
	fread(pic, len, 1, fp);
	fclose(fp);
		
	if (strncmp("**TI83**", (char *)pic, 8) != 0) {
		printf("Invalid TI83 File!\n");
		return 0;
	}
	
	ptr = pic;
	ptr += 11; /* file type ? */
	ptr += 0x2A; /* comment block */
	
	len2 = *ptr | (*(ptr + 1) << 8);
	if (len != (len2 + 57)) {
		printf("Multiple files detected! Only first one can be used...\n");
	}
		
	ptr += 2; /* file length */
	ptr += 2; /* 11 */
	ptr += 2; /* length of data */
	
	if (*ptr != 0x7) {
		printf("File is not a TI83 Pic!\n");
		return 0;
	}
	
	ptr += 1; /* type */ 
	ptr += 8; /* name */
	ptr += 2; /* length again ? */
	ptr += 2; /* length - 2 ? */

	fp = fopen(inc_file, "w");
	for (x = 0; x < 63; x++) {
		fprintf(fp, ".db ");
		for (y = 0; y < (96 / 8); y++) {
			fprintf(fp, "%d,", *ptr);
			ptr++;
		}
		fprintf(fp, "\n");
	}	
	fclose(fp);
	
	free(pic);
	
	return 1;
}

int pbm_to_pic(char *pbm_file, char *pic_file, unsigned char var)
{
	signed char sig[] = {'*', '*', 'T', 'I', '8', '3', '*', '*', 0x1A, 0x0A, 0x0};
	
	FILE *fp;
	
	int cols, rows;
	int x, y, z;
	
	bit **pbm_data;
	unsigned char *pic_data;
	unsigned short int chksum;
	
	fp = fopen(pbm_file, "r");
	pbm_data = pbm_readpbm(fp, &cols, &rows);
	fclose(fp);
	
	if ((cols != 96) || (rows != 63)) {
		printf("File not valid size for PIC!\n");
		return 0;
	}
	
	pic_data = (unsigned char *)malloc((96*63)/8);
	
	for (x = 0; x < rows; x++) {
		for (y = 0; y < (cols / 8); y++) {
			for (z = 0; z < 8; z++) {
				pic_data[((96/8) * x) + y] |= (pbm_data[x][(y * 8) + z]) << (7 - z);
			}
		}
	}
	
	fp = fopen(pic_file, "w");
	
	fwrite(sig, sizeof(sig), 1, fp);
	
	for (x = 0; x < COMMENT_LENGTH; x++)
		fputc(0x0, fp);
	
	chksum = 0x305; 
	fputc((chksum & 0x00FF), fp);
	fputc((chksum >> 8), fp);
	
	chksum = 0;
	
	x = 11;
	fputc((x & 0x00FF), fp); chksum += (x & 0x00FF);
	fputc((x >> 8), fp); chksum += (x >> 8);
	
	x = (63*96)/8 + 2;
	fputc((x & 0x00FF), fp); chksum += (x & 0x00FF);
	fputc((x >> 8), fp); chksum += (x >> 8);
	
	fputc(0x7, fp);	chksum += 0x7;
	
	fputc(0x60, fp); chksum += 0x60;
	fputc(var, fp); chksum += var;
	for (x = 2; x < 8; x++)
		fputc(0x0, fp);
	
	x = (63*96)/8 + 2;
	fputc((x & 0x00FF), fp); chksum += (x & 0x00FF);
	fputc((x >> 8), fp); chksum += (x >> 8);
	
	x = (63*96)/8;
	fputc((x & 0x00FF), fp); chksum += (x & 0x00FF);
	fputc((x >> 8), fp); chksum += (x >> 8);
	
	for (x = 0; x < ((63*96)/8); x++) {
		fputc(pic_data[x], fp);
		chksum += pic_data[x];
	}
	
	fputc((chksum & 0x00FF), fp);
	fputc((chksum >> 8), fp);
	
	fclose(fp);
	free(pic_data);
	
	pbm_freearray(pbm_data, 63);
		
	return 1;
}

int pbm_to_inc(char *pbm_file, char *inc_file)
{
	FILE *fp;
	
	int cols, rows;
	int x, y, z;
	int dat;
	
	bit **pbm_data;
	
	fp = fopen(pbm_file, "r");
	pbm_data = pbm_readpbm(fp, &cols, &rows);
	fclose(fp);
	
	fp = fopen(inc_file, "w");
	
	for (x = 0; x < rows; x++) {
		fprintf(fp, ".db ");
		for (y = 0; y < (cols / 8); y++) {
			dat = 0;
			for (z = 0; z < 8; z++) {	
				dat |= pbm_data[x][(y * 8) + z] << (7 - z);
			}
			fprintf(fp, "%d,", dat);
		}
		if (cols % 8) {
			dat = 0;
			for (z = (cols % 8); z < 8; z++) {
				dat |= pbm_data[x][(y * 8) + z] << (7 - z);
			}
			fprintf(fp, "%d,", dat);
		}
		fprintf(fp, "\n");
	}
	
	fclose(fp);
	
	return 1;
}

int main(int argc, char *argv[])
{
	int opt;
	unsigned char var = 0;
	
	while ((opt = getopt(argc, argv, "hv:")) != -1) {
		switch(opt) {
			case 'h':
				usage(argv[0]);
				exit(EXIT_SUCCESS);
			case 'v':
				var = atoi(optarg);
				if (var > 10) {
					fprintf(stderr, "83I variable number is invalid\n");
					exit(EXIT_FAILURE);
				}
				break;
			case ':':
				fprintf(stderr, "option needs a value\n");
				exit(EXIT_FAILURE);
			case '?':
				fprintf(stderr, "unknown option: %c\n", optopt);
				exit(EXIT_FAILURE);
		}
	}
	
	if ((argc - optind) != 2) {
		fprintf(stderr, "Incorrect amount of parameters, Try %s -h\n", argv[0]);
		exit(EXIT_FAILURE);
	}
	if (strstr(argv[optind], ".83i")) {
		if (strstr(argv[optind + 1], ".pbm")) {
			pic_to_pbm(argv[optind], argv[optind + 1]);
		} else if (strstr(argv[optind + 1], ".inc")) {
			pic_to_inc(argv[optind], argv[optind + 1]);
		} else {
			fprintf(stderr, "Invalid output type\n");
		}
	} else if (strstr(argv[optind], ".pbm")) {
		if (strstr(argv[optind + 1], ".83i")) {
			pbm_to_pic(argv[optind], argv[optind + 1], var);
		} else if (strstr(argv[optind + 1], ".inc")) {
			pbm_to_inc(argv[optind], argv[optind + 1]);
		} else {
			fprintf(stderr, "Invalid output type\n");
		}
	} else {
		fprintf(stderr, "Invalid input type\n");
	}
	
	return 0;
}
