/*
 *  titranz.c - ti83 transmit program
 *  Copyright (C) 1998
 *
 *  Original Author: Steven Fuller <relnev@netscape.net>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>
#include <pnm.h>

short lpt_out = 0x378;
short lpt_in  = 0x379;

int mode = -1;
int calc = 0;

#define SLEEP_RATE 1

#define MODE_SEND 0
#define MODE_RECV 1
#define MODE_ROMDUMP 2
#define MODE_SCREENDUMP 3
#define MODE_BACKUPSEND 4
#define MODE_BACKUPRECV 5

#define TI83 3

#define TI83_ROMSIZE 262144

#define getbyte paraget
#define putbyte paraput

static void outb (unsigned short int port, unsigned char val)
{
  __asm__ volatile (
  	"outb %0,%1\n"
  	:
  	: "a" (val), "d" (port)
  	);
}

static int inb (short port)
{
  unsigned char val;
  __asm__ volatile ("inb %1,%0"
  		    : "=a" (val)
  		    : "d" ((unsigned short)port));
  return val;
}

void init_link(void)
{
	if (ioperm(lpt_out,1,1) || ioperm(lpt_in,1,1))
	{
		fprintf(stderr, "Cannot access ports 0x%x and 0x%x, aborting!\n", lpt_out, lpt_in);
		exit(EXIT_FAILURE);
	}
	outb(lpt_out, 3);
}

/* paraput/paraget/init_link code from xlink85 */
               
/* send a byte thru parallel link */
void paraput(unsigned char data)
{
	int bit;
	/* printf(" {0x%x}\n ", data); */
	for (bit=0; bit<8; bit++) {
		if (data&1) {
			outb(lpt_out, 2);
			while (inb(lpt_in)&0x10);
			outb(lpt_out, 3);
			while ((inb(lpt_in)&0x10)==0x00);
		} else {
			outb(lpt_out, 1);
			while (inb(lpt_in)&0x20);
			outb(lpt_out, 3);
			while ((inb(lpt_in)&0x20)==0x00);
		}
		data>>=1;
	}
}

int timetest() 
{
	/* timeout? */
 	return 0;
}

/* receive a byte thru parallel link */
unsigned char paraget(void) 
{
	int bit;
	unsigned char data=0, v;

	for (bit=0; bit<8; bit++) {
	        while ((v=inb(lpt_in)&0x30) == 0x30) timetest();
		if (v==0x10) {
			data=(data>>1)|0x80;
			outb(lpt_out, 1);
		        while ((inb(lpt_in)&0x20)==0x00) timetest();
			outb(lpt_out, 3);
		} else {
			data=data>>1;
			outb(lpt_out, 2);
			while ((inb(lpt_in)&0x10)==0x00) timetest();
			outb(lpt_out, 3);
		}
	}
	usleep(SLEEP_RATE);
	/* printf(" [0x%x]\n ", data); */
	return data;
}

void usage(char *s)
{
	printf("\n");
	printf("%s - ti83 transmit program\n", s);
	printf("by Steven Fuller <relnev@netscape.net>\n");
	printf("\tusage: %s <files>\n", s);
	printf("Optional parameters:\n");
	printf("\t-d <filename>: wait and recieve romdump (rom83.83p)\n");
	printf("\t-h: this text\n");
	printf("\t-s <filename>: get and save screenshot\n");
	printf("\n");
}

int main (int argc, char *argv[])
{
	FILE *fp;
	char *filename = NULL;
	unsigned char *buf, *ptr;
	int opt;
	int x, y, z, filelen, length;
	int done;
	unsigned short chksum, chktmp, tlength;
	unsigned short paklen, datalen;
	unsigned char namelen, datatype;
	unsigned char *dataname = NULL;
	bit **pbm_data;
	
	mode = MODE_SEND;
	
	while ((opt = getopt(argc, argv, "d:hs:")) != -1) {
		switch(opt) {
		case 'd':
			mode = MODE_ROMDUMP;
			filename = strdup(optarg);
			break;
		case 'h':
			usage(argv[0]);
			exit(EXIT_SUCCESS);
		case 's':
			mode = MODE_SCREENDUMP;
			filename = strdup(optarg);
			break;
		case ':':
			printf("option needs a value\n");
			exit(EXIT_FAILURE);
		case '?':
			printf("unknown option: %c\n", optopt);
			exit(EXIT_FAILURE);
		}
	}

	if (mode == -1) {
		mode = MODE_RECV;
	}	
			
	init_link();
	
	if (mode == MODE_ROMDUMP) {
		fp = fopen(filename, "wb");
		for (x = 0; x < TI83_ROMSIZE; x++) {
			if ((x % 256) == 0)
				printf("%d\n", x);
			fputc(getbyte(), fp);
		}
		fclose(fp);
	} else if (mode == MODE_SCREENDUMP) {
		putbyte(0x3);
		putbyte(0x6D);
		putbyte(0x0);
		putbyte(0x0);
		getbyte();
		getbyte();
		getbyte();
		getbyte();
		getbyte();
		getbyte();
		getbyte();
		getbyte();
		buf = (unsigned char *)malloc(768);
		for (x = 0; x < 768; x++)
			*(buf + x) = getbyte();
		getbyte();
		getbyte();
		putbyte(0x3);
		putbyte(0x56);
		putbyte(0x0);
		putbyte(0x0);
		ptr = buf;
		pbm_data = (bit **)malloc(sizeof(bit *) * 64);
		for (x = 0; x < 64; x++) {
			pbm_data[x] = (bit *)calloc(sizeof(bit *) * 96, 1);
			for (y = 0; y < (96 / 8); y++) {
				for (z = 0; z < 8; z++)
					pbm_data[x][(y * 8) + z] = (*ptr & (1 << (7 - z))) >> (7 - z);
				ptr++;
			}
		}
		fp = fopen(filename, "wb");
		pbm_writepbm(fp, (bit **)pbm_data, 96, 64, 0);
		fclose(fp);
		for (x = 0; x < 64; x++)
			free(pbm_data[x]);
		free(pbm_data);
		free(buf);
	} else if (mode == MODE_SEND) {
		tlength = 0;
		for(; optind < argc; optind++) {
			if (!(fp = fopen(argv[optind], "r"))) {
				printf("File %s could not be loaded!\n", argv[optind]);
				continue;
			}
			fseek(fp, 0, SEEK_END);
			filelen = ftell(fp);
			rewind(fp);
			buf = (unsigned char *)malloc(filelen);
			fread(buf, filelen, 1, fp);
			fclose(fp);
			if (!((*buf == '*') || (*(buf + 1) == '*') || 
			     (*(buf + 2) == 'T') || (*(buf + 3) == 'I') ||
			     (*(buf + 4) == '8') || (*(buf + 6) == '*') ||
			     (*(buf + 7) == '*'))) {
			     	printf("%s is not a TI File!\n", argv[optind]);
			     	continue;
			}
			calc = *(buf + 5) - '0';
			switch(calc) {
			case TI83:
				break;
			default:
				printf("%s requires an unsupported calculator!\n", argv[optind]);
				continue;
			}
			ptr = buf + 0x35;
			length = *ptr | (*(ptr + 1) << 8);
			ptr += 2; 
			if ((length + 0x35 + 4) != filelen) {
				printf("%s is the incorrect length!\n", argv[optind]);
				continue;
			}
			chksum = *(buf + filelen - 2) | (*(buf + filelen - 1) << 8);
			chktmp = 0;
			for (x = 0; x < length; x++) {
				chktmp += *(ptr + x);
			}
			if (chktmp != chksum) {
				printf("%s has an invalid checksum 0x%X, should be 0x%X!, %d\n", argv[optind], chktmp, chksum, length);
				continue;
			}
			done = 0;
			printf("Sending file %s...\n", argv[optind]);
			while (!done) {
				paklen = *ptr | (*(ptr + 1) << 8); /* packet length */
				chksum = 0;
				ptr += 2; 
				datalen = *(ptr + paklen) | (*(ptr + paklen + 1) << 8);
				chksum += *(ptr + paklen); /* needed for easy byte creation */
				chksum += *(ptr + paklen + 1);
				ptr += 2;
				datatype = *ptr;
				chksum += *ptr;
				ptr += 1;
				if (dataname) free(dataname);
				if (calc == 0x5) { /* I dunno what else would need this? */
					namelen = *ptr;
					ptr++;
				} else {
					namelen = paklen - sizeof(datalen) - sizeof(datatype);
				}
				dataname = malloc(namelen + 1);
				memset(dataname, '\0', namelen + 1);
				for (x = 0; x < namelen; x++) {
					*(dataname + x) = *(ptr + x);
					chksum += *(ptr + x);
				}
				printf("Outputting Variable with type 0x%x with name %s.\n", datatype, dataname);
				ptr += namelen;
				if ((ptr + datalen + 3) == (buf + filelen - 1)) {
					/* check to see if this file is done with */
					done = 1; 
				}
				ptr += 2;
				for (x = 0; x < 3; x++) { 
					/* lets give it three tries */
					putbyte(calc); /* computer id */
					putbyte(0x6); /* var header cmd */
					putbyte(paklen & 0x00FF);
					putbyte(paklen >> 8);
					printf("[0x%x] - paklen\n", paklen);
					putbyte(datalen & 0x00FF);
					putbyte(datalen >> 8);
					printf("[0x%x] - datalen\n", datalen);
					putbyte(datatype);
					printf("[0x%x] - datatype\n", datatype);
					if (calc == 0x5) {
						/* wonder what other calcs want this */
						putbyte(namelen);
					}
					printf("[");
					for (opt = 0; opt < namelen; opt++) {
						putbyte(*(dataname + opt));
						printf("%c", *(dataname + opt));
					}
					printf("]");
					putbyte(chksum & 0x00FF);
					putbyte(chksum >> 8);
					if (getbyte() != (calc + 0x80)) {
						/* done = 1; */
						/* i guess go with it? */
						printf("Whoa now? Different calc?\n");
						/* continue; */
					}
					getbyte(); getbyte(); getbyte(); /* 0x56 0x0 0x0 */
					getbyte(); /* we know what calc this is */
					switch (getbyte()) {
						case 0x9:
							/* onward */
							getbyte(); getbyte();
							x = 3;
							continue;
						case 0x5A:
							printf("Checksum error: Retry %d.\n", x + 2);
							getbyte(); getbyte();
							break;
						case 0x36:
							getbyte(); getbyte();
							switch (getbyte()) {
								case 0x1:
									getbyte(); getbyte();
									/* I presume to exit? */
								case 0x3:
									exit(EXIT_SUCCESS);
								case 0x2:
									getbyte(); getbyte();
									x = 3;
									break;
							}
					}						
				}
				/* if (x == 3) continue; */
				putbyte(calc);
				putbyte(0x56); /* ok */
				putbyte(0); putbyte(0);
				printf("?\n");
				for (x = 0; x < 3; x++) {
					putbyte(calc);
					putbyte(0x15);
					putbyte(datalen & 0x00FF);
					putbyte(datalen >> 8);
					chksum = 0;
					for (x = 0; x < datalen; x++) {
						chksum += *(ptr + x);
						putbyte(*(ptr + x));
					}
					putbyte(chksum & 0x00FF);
					putbyte(chksum >> 8);
					getbyte(); 
					switch(getbyte()) {
					case 0x56:
						getbyte(); getbyte();
						x = 3;
						continue;
					case 0x5A:
						printf("Checksum error: Retry %d.\n", x + 2);
						getbyte(); getbyte();
					}
				}
				/* if (x == 3) continue; */
				ptr += datalen;
				tlength += datalen;
				if (done && ((argc - 1) == optind)) {
					putbyte(calc);
					putbyte(0x92); /* no more */
					printf("Transferred %d bytes.\n", tlength);
					putbyte(tlength & 0x00FF);
					putbyte(tlength >> 8);
					getbyte(); getbyte();
					getbyte(); getbyte();
				}
			}
			free(buf);
		}
		printf("All done!\n");
	}
	return 0;
}
